/* pybitmap: a bitmap extension class for Python.
   
   Copyright (C) 2000 Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

   I wrote this class to play around with Python's C Extension
   API.  As it is, I'm pretty happy that it works, but I'm sure
   that some parts look very amateurish.  I want to learn the
   extension API a bit better, and would be very happy if people could
   email improvements to me.

   A bitmap is a classic data structure that acts like a long array of
   switches.  The advantage of the bitmap versus a regular Python list
   is in space savings: each "switch" takes only one bit of memory.
   That is, we can make a bitmap of 1 billion switches for around 125
   megabytes.  In comparison, the equivalent Python list would take at
   least 32*125 megabytes = 4 gigs, so we're saving a lot of memory.  */

#include "Python.h"
#include "bitmap.h"

/* Later, I might remove the dependency on bitmap.c/h and bundle the
   functionality directly into pybitmapmodule.c. */

/* Definition of a new PyObject type --- this Bitmap object contains
 pointers to a new Bitmap. */
typedef struct {
  PyObject_HEAD
  Bitmap b;
} PyBitmapobject;


/* **********************************************************************/
/* Here's a list of prototypes so that C isn't too baffled: */
static PyBitmapobject* PyBitmap_alloc(PyObject *module, PyObject *args);
static void PyBitmap_dealloc(PyBitmapobject *self);
static PyObject* PyBitmap_getattr(PyBitmapobject *self, char *name);
static PyObject* PyBitmap_set(PyBitmapobject *self, PyObject *args);
static PyObject* PyBitmap_unset(PyBitmapobject *self, PyObject *args);
static PyObject* PyBitmap_get(PyBitmapobject *self, PyObject *args);

/* Sequence functions, which are easier to use (and bounds checked for
   safety). */
static int PyBitmap_length(PyBitmapobject *self);
static PyObject* PyBitmap_item(PyBitmapobject *self, int i);
static int PyBitmap_ass_item(PyBitmapobject *self, int i, 
				   PyObject *value);

staticforward char PyBitmap_set_doc[];
staticforward char PyBitmap_unset_doc[];
staticforward char PyBitmap_get_doc[];
/* Question: What is the staticforward keyword?  Later on, we'll be
   defining PyBitmaptype, but I wonder if it's ok to move the
   declaration up here instead.  Better break out K&R.  Nope, it's not
   defined in there.  It must be something custom to the Python
   system.

   Ah!  It's something that the Python people have defined in
   Include/object.h.  It handles compiler oddities to forward
   declarations of static things.  Otherwise, it's the same as the
   "static" keyword.  */
staticforward PyTypeObject PyBitmaptype;
staticforward PySequenceMethods PyBitmap_as_sequence;

static char PyBitmaptype_doc[] = "The PyBitmap type.";
/* Let's define everything we can do with a PyBitmap.*/
static PyTypeObject PyBitmaptype = {
  PyObject_HEAD_INIT(NULL) /* see FAQ 3.24 for explanation */
  0,				/*ob_size*/
  "PyBitmap",			/*tp_name*/
  sizeof(PyBitmapobject),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)PyBitmap_dealloc,	/*tp_dealloc*/
  (printfunc)0,		/*tp_print*/
  (getattrfunc)PyBitmap_getattr,	/*tp_getattr*/
  (setattrfunc)0,	/*tp_setattr*/
  (cmpfunc)0,		/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,			/*tp_as_number*/
  &PyBitmap_as_sequence,  /*tp_as_sequence --- see PyBitmap_as_sequence.*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,		/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  
  /* Space for future expansion */
  0L,0L,0L,0L,
  PyBitmaptype_doc /* Documentation string */
};

/* PyBitmaps will support a limited set of sequence functions.  I've
   taken a look at Objects/listobject.c and mimicked it. */
static PySequenceMethods PyBitmap_as_sequence = {
        (inquiry)PyBitmap_length, /*sq_length*/
        (binaryfunc)0, /*sq_concat*/
        (intargfunc)0, /*sq_repeat*/
        (intargfunc)PyBitmap_item, /*sq_item*/
        (intintargfunc)0, /*sq_slice*/
        (intobjargproc)PyBitmap_ass_item, /*sq_ass_item*/
        (intintobjargproc)0, /*sq_ass_slice*/
        (objobjproc)0, /*sq_contains*/
        (binaryfunc)0, /*sq_inplace_concat*/
        (intargfunc)0, /*sq_inplace_repeat*/
};


/* **********************************************************************/
/* Method definitions */

/* Constructing a new PyBitmap */
static char PyBitmap_alloc_doc[] = "Construct a new PyBitmap of size n.";
static PyBitmapobject* PyBitmap_alloc(PyObject *module, PyObject *args) {
  PyBitmapobject *self = NULL;
  long size = 0;
  /* fixme --- I should bounds check to make sure they don't give a
     negative size. */
  if(PyArg_ParseTuple(args, "l", &size)) {
    self = PyObject_New(PyBitmapobject, &PyBitmaptype);
    self->b = Bitmap_new(size);
  }
  return self;
}

/* Destructing a bitmap */
static void PyBitmap_dealloc(PyBitmapobject *self) {
  Bitmap_delete(self->b);
  PyObject_Del(self);
}

/* Let's tell Python which methods a PyBitmap can support.*/
static struct PyMethodDef PyBitmap_methods[] = {
  {"set", (PyCFunction)PyBitmap_set, METH_VARARGS, 
   PyBitmap_set_doc},
  {"unset", (PyCFunction)PyBitmap_unset, METH_VARARGS, 
   PyBitmap_unset_doc},
  {"get", (PyCFunction)PyBitmap_get, METH_VARARGS,
   PyBitmap_get_doc},
  {NULL, NULL} /* Sentinel. */
};
/* And here's our version of the __getattr__ method.*/
static PyObject *
PyBitmap_getattr(PyBitmapobject *self, char *name) {
  return Py_FindMethod(PyBitmap_methods, (PyObject *)self, name);
}

static char PyBitmap_set_doc[] = 
"Turn on bit n.\n\
\n\
This is not bounds checked, so you might want to use the regular\n\
element assignment operator instead.\n\
\n\
Sample usage:\n\
    b.set(0)\n\
";
static PyObject* PyBitmap_set(PyBitmapobject *self, PyObject *args) {
  long index;
  if (PyArg_ParseTuple(args, "l", &index)) {
    Bitmap_set(self->b, index);
  }
  else return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}

static char PyBitmap_unset_doc[] = 
"Turn off bit n.\n\
\n\
This is not bounds checked, so you might want to use the regular\n\
element assignment operator instead.\n\
\n\
Sample usage:\n\
    b.unset(0)\n\
";
static PyObject* PyBitmap_unset(PyBitmapobject *self, PyObject *args) {
  long index;
  if (PyArg_ParseTuple(args, "l", &index)) {
    Bitmap_unset(self->b, index);
  }
  else return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}

static char PyBitmap_get_doc[] = 
"Look up the value at bit n.\n\
\n\
This is not bounds checked, so you might want to use the regular\n\
element look up operator instead.\n\
\n\
Sample usage:\n\
    print b.get(0)\n\
";
static PyObject* PyBitmap_get(PyBitmapobject *self, PyObject *args) {
  long index;
  int value;
  if(PyArg_ParseTuple(args, "l", &index)) {
    value = Bitmap_get(self->b, index);
    return PyInt_FromLong(value);
    /* fixme: I should do some bounds checking... Let's do that later. */
  }
  else return NULL;
}

/* Return the size of a PyBitmap. */
static int PyBitmap_length(PyBitmapobject *self) {
  return Bitmap_getSize(self->b);
}

/* PyBitmap_item and PyBitmap_ass_item will do bounds checking, so
   it'll be safer to use. */
static PyObject* PyBitmap_item(PyBitmapobject *self, int i) {
  if(i < 0 || i >= Bitmap_getSize(self->b)) {
    PyErr_SetString(PyExc_IndexError, "bitmap index out of range.");
    return NULL; 
  }
  return PyInt_FromLong(Bitmap_get(self->b, i));
}

/* The way that sequence assignment works is a little weirder --- we
   need to return 1 if things broke, and 0 otherwise.  It's a bit
   weird.  This assignment is bounds checked. */
static int PyBitmap_ass_item(PyBitmapobject *self, int i, 
				   PyObject *value) {
  if(i < 0 || i >= Bitmap_getSize(self->b)) {
    PyErr_SetString(PyExc_IndexError, "bitmap index out of range.");
    return 1; /* Return 1 on failure */
  }
  if (PyInt_AsLong(value))
    Bitmap_set(self->b, i);
  else 
    Bitmap_unset(self->b, i);
  return 0; /* Return 0 on success */
}



/* ***********************************************************************/
/* Module initialization stuff */

static char pybitmap_module_doc[] = 
"This module provides a bitmap structure called PyBitmap.\n\
\n\
Sample use:\n\
\n\
    from pybitmap import PyBitmap\n\
    mybitmap = PyBitmap(100)\n\
    for i in range(len(mybitmap)):\n\
        print mybitmap[i],\n\
        mybitmap[i] = 1\n\
        print mybitmap[i],\n\
        mybitmap[i] = 0\n\
        print mybitmap[i]\n\
\n\
PyBitmaps support the indexing and element assignment operators of\n\
sequences, and these functions are bounds checked.  Also, there are\n\
other functions that aren't bounds checked:\n\
\n\
    set(self, n)\n\
    unset(self, n)\n\
    get(self, n)\n\
";
/* The only "function" we'll define for the pybitmap module is the
   PyBitmap() constructor. */
static struct PyMethodDef pybitmap_methods[] = {
  {"PyBitmap", (PyCFunction)PyBitmap_alloc, METH_VARARGS, 
   PyBitmap_alloc_doc},
  {NULL, (PyCFunction)NULL, 0, NULL}		/* sentinel */
};

void initpybitmap(void) {
  PyObject *m, *d;
  /* It's necessary to manually set the type like this to support
     Windows platforms.  Check Python FAQ entry 3.24 for an
     explanation. */
  PyBitmaptype.ob_type = &PyType_Type; 
  /* Create the module and add the functions */
  m = Py_InitModule4("pybitmap", pybitmap_methods,
		     pybitmap_module_doc,
		     (PyObject*)NULL,PYTHON_API_VERSION);
  /* Add some symbolic constants to the module */
  d = PyModule_GetDict(m);
  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module pybitmap.");
}
